import pandas as pd

from earningspy.common.constants import (
    FINVIZ_EARNINGS_DATE_KEY,
    DAYS_TO_EARNINGS_KEY_CAPITAL,
    TICKER_KEY_CAPITAL,
    ABS_RET_KEY,
    EXP_RET_KEY,
    RF_KEY,
    MARK_EXP_KEY,
    CAPM_KEY,
    VIX_KEY,
    EARNING_VIX_KEY,
    CAR_KEY,
    BHAR_KEY,
    ALLOWED_WINDOWS,
    AVAILABLE_CHECK_COLUMNS,
)
from earningspy.calendars.utils import days_left
from earningspy.inspectors.mixins import CARMixin, TimeSeriesMixin


class PEADInspector(CARMixin, TimeSeriesMixin):
    """
    Calendar should be passed using this method
    pd.read_csv('<field-name>.csv', index_col=0, parse_dates=True)
    """

    def __init__(self, 
                 calendar=None,
                 price_history=None):

        self.calendar = self._load_calendar(calendar)
        self.backup = self.calendar.copy()
        self.price_history = self._load_price_history(price_history)

        self.remaining_data = self.calendar[~(self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] < -3)]
        self.merged_data = None


    def _load_calendar(self, calendar):

        calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] = calendar.apply(lambda row: days_left(row), axis=1)
        calendar = calendar.sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)

        return calendar
    
    def _sort_calendar(self):
        self.calendar = self.calendar.sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)


    def inspect(self, days=3, dry_run=False, reuse_timeseries=False, post_earnings=False, async_=False):

        if days not in ALLOWED_WINDOWS:
            raise Exception(f'Invalid day range. Select from {ALLOWED_WINDOWS}')

        self.affected_rows = self._get_affected_rows(days, post_earnings=post_earnings)
        if dry_run:
            self.affected_rows = self.affected_rows.reset_index()
            self.affected_rows = self.affected_rows.set_index([FINVIZ_EARNINGS_DATE_KEY])
            return self.affected_rows

        self._process_windows_columns(days=days, reuse_timeseries=reuse_timeseries, async_=async_)
        self._get_earnings_vix()
        self._find_and_remove_duplicates()
        self._sort_calendar()

        return self

    def refresh(self, days=3, dry_run=False, reuse_timeseries=False, check_column=None, deep=False, async_=False):

        if days not in ALLOWED_WINDOWS:
            raise Exception(f'Invalid day range. Select from {ALLOWED_WINDOWS}')

        if not check_column or check_column not in AVAILABLE_CHECK_COLUMNS:
            raise Exception(f"Provide a column to check for NaNs to do the refresh, must be from this list {AVAILABLE_CHECK_COLUMNS}")
        
        self.affected_rows = self._get_affected_rows(days, check_column=check_column, deep=deep)
        if dry_run:
            self.affected_rows = self.affected_rows.reset_index()
            self.affected_rows = self.affected_rows.set_index([FINVIZ_EARNINGS_DATE_KEY])
            return self.affected_rows
        
        self._process_windows_columns(days=days, reuse_timeseries=reuse_timeseries, async_=async_)
        self._get_earnings_vix()
        self._find_and_remove_duplicates()

        return self
        

    def _process_windows_columns(self, days=3, reuse_timeseries=False, async_=False):

        if not reuse_timeseries:
            self.price_history = self.fetch_price_history(assets=set(self.affected_rows.index.get_level_values(1).to_list()), async_=async_)

        self.calendar = self.calendar.reset_index()
        self.calendar = self.calendar.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])

        self._get_windows_abnormal_return(days=days)
        self._get_windows_risk_free_rate(days=days)
        self._get_windows_expected_return(days=days)
        self._get_windows_market_expected_return(days=days)
        self._get_windows_capm(days=days)
        self._get_windows_car(days=days)
        self._get_windows_bhar(days=days)
        self._get_windows_vix(days=days)


    def _get_affected_rows(self, days, check_column=ABS_RET_KEY, deep=False, post_earnings=False):

        start = days
        end = days + 30

        if deep:
            affected_rows = self.calendar[(self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] <= -start)]
        else:
            affected_rows = self.calendar[(self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] <= -start) &
                                          (self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] >= -end)]

        if not post_earnings:
            try:
                affected_rows = affected_rows[affected_rows[check_column.format(days)].isna()]
            except KeyError:
                raise Exception(f"Check column {check_column.format(days)} not found in the calendar. Is this post earnings data?"
                       " If so, set post_earnings=True.")
        affected_rows = affected_rows.reset_index()
        affected_rows = affected_rows.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])
        return affected_rows


    def join(self, storage, earnings_phase: str = "pre", keep="last"):
        """
        Join calendar with storage data.

        :param storage: DataFrame with historical data
        :param earnings_phase: 'pre' or 'post'
        :param keep: 'first' or 'last'
            first -> preserve data from new storage values
            last -> Preserves data from the canonical storage (default)
        """

        if storage is None or storage.empty:
            raise Exception("storage can't be empty")

        if self.calendar.empty:
            raise Exception("calendar is empty nothing to concat")

        if earnings_phase not in {"pre", "post"}:
            raise ValueError("earnings_phase must be either 'pre' or 'post'")

        if keep not in {"first", "last"}:
            raise ValueError("Keep must be either 'first' or 'last'")

        if earnings_phase == "pre":
            # Accounts for BMO row cases.
            storage = storage[storage[DAYS_TO_EARNINGS_KEY_CAPITAL] > 0]
        else: 
            # Account for AMC row cases.
            storage = storage[storage[DAYS_TO_EARNINGS_KEY_CAPITAL] <= -1]

        self.merged_data = (
            pd.concat([storage, self.calendar], join="outer")
            .sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)
            .reset_index()
            .drop_duplicates(subset=[FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL], keep=keep)
            .set_index(FINVIZ_EARNINGS_DATE_KEY)
        )

        self.merged_data[DAYS_TO_EARNINGS_KEY_CAPITAL] = (
            self.merged_data.apply(lambda row: days_left(row), axis=1)
        )
        self.merged_data = self.merged_data.sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)

        return self.merged_data


    def _get_windows_abnormal_return(self, days):

        label = ABS_RET_KEY.format(days)
    
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_window_pct_change(row, days=days), axis=1)

    
    def _get_windows_market_expected_return(self, days):
        
        label = MARK_EXP_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_market_expected_return(row, days=days), axis=1)


    def _get_windows_capm(self, days):

        label = CAPM_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_capm(row, days=days), axis=1)


    def _get_windows_expected_return(self, days):

        label = EXP_RET_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_expected_return(row, days=days), axis=1)


    def _get_windows_risk_free_rate(self, days):
        label = RF_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_risk_free_rate(row, days=days), axis=1)


    def _get_windows_vix(self, days):

        label = VIX_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_vix(row, days=days), axis=1)


    def _get_earnings_vix(self):

        self.calendar.loc[self.affected_rows.index, EARNING_VIX_KEY] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_vix_for_date(row), axis=1)


    def _get_windows_car(self, days):
        label = CAR_KEY.format(days)
        ret_label = ABS_RET_KEY.format(days)
        capm_label = CAPM_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = (self.calendar[ret_label] - self.calendar[capm_label]).round(4)


    def _get_windows_bhar(self, days):
        label = BHAR_KEY.format(days)
        ret_label = ABS_RET_KEY.format(days)
        benchmark_label = MARK_EXP_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = (self.calendar[ret_label] - self.calendar[benchmark_label]).round(4)


    def _find_and_remove_duplicates(self):
        self.calendar = self.calendar.reset_index()
        self.calendar = self.calendar.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])
        self.calendar = self.calendar[~self.calendar.index.duplicated(keep='first')]

        self._remove_duplicate_ticker_quarters()

        self.calendar = self.calendar.reset_index()
        self.calendar = self.calendar.set_index(FINVIZ_EARNINGS_DATE_KEY)


    def _remove_duplicate_ticker_quarters(self):
        self.calendar = self.calendar.reset_index()
        self.calendar['year'] = self.calendar[FINVIZ_EARNINGS_DATE_KEY].dt.year
        self.calendar['quarter'] = self.calendar[FINVIZ_EARNINGS_DATE_KEY].dt.quarter
        self.calendar = self.calendar.sort_values(FINVIZ_EARNINGS_DATE_KEY)
        original_len = len(self.calendar)
        self.calendar = self.calendar.drop_duplicates(subset=[TICKER_KEY_CAPITAL, 'year', 'quarter'], keep='last')
        num_duplicates = original_len - len(self.calendar)
        print(f"Found and removed {num_duplicates} duplicate ticker-quarter entries.")
        self.calendar = self.calendar.drop(columns=['year', 'quarter'])
        self.calendar = self.calendar.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])



import asyncio
import aiohttp
import random

from datetime import datetime as dt
from dateutil.relativedelta import relativedelta
import pandas as pd
from tqdm import tqdm

# Existing functions (get_range, get_range_timestamps, prepare_data) remain unchanged
def get_range(range_from, end_date):

    accepted_values = ['3m', '9m', '1y', '5y', '10y']
    if range_from not in accepted_values:
        raise Exception('Invalid from value')

    if range_from == '3m':
        start_date = end_date - relativedelta(months=3)
    if range_from == '9m':
        start_date = end_date - relativedelta(months=9)
    if range_from == '1y':
        start_date = end_date - relativedelta(years=1)
    if range_from == '5y':
        start_date = end_date - relativedelta(years=5)
    if range_from == '10y':
        start_date = end_date - relativedelta(years=10)
    
    return str(start_date), str(end_date)


def get_range_timestamps(start_date, end_date):

    start_date = str(dt.strptime(start_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    end_date = str(dt.strptime(end_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    return start_date, end_date

async def get_one_ticker_async(session, asset, from_='3m', start_date=None, end_date=dt.now().date()):
    headers = {
        'User-Agent': "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; .NET CLR 1.0.3705;)",
    }

    if start_date:
        end_date = str(end_date)
        start, end = get_range_timestamps(start_date, end_date)
    else:
        start, end = get_range(range_from=from_, end_date=end_date)
        start, end = get_range_timestamps(start, end)

    url = f"https://query2.finance.yahoo.com/v8/finance/chart/{asset}?period1={start}&period2={end}&interval=1d&events=history"
    
    try:
        async with session.get(url, headers=headers) as response:
            if response.status == 200:
                data = await response.json()
                df = pd.DataFrame.from_dict(data['chart']['result'][0]['indicators']['quote'][0])
                df['Date'] = data['chart']['result'][0]['timestamp']
                df['Date'] = df['Date'].apply(dt.fromtimestamp)
                df = df.set_index('Date', drop=True)
                df.index = df.index.normalize()
                return asset, df.round(2)
            else:
                # print(f"Could not retrieve data for {asset}: {response.status}")
                return asset, None
    except Exception as e:
        print(f"Error fetching data for {asset}: {str(e)}")
        return asset, None

async def get_portfolio_async(assets, from_='3m', start_date=None, end_date=dt.now().date()):
    portfolio = pd.DataFrame()
    not_found = []

    async with aiohttp.ClientSession() as session:
        tasks = []
        for asset in assets:
            tasks.append(asyncio.ensure_future(get_one_ticker_async(session, asset, from_=from_, start_date=start_date, end_date=end_date)))
        
        for future in tqdm(asyncio.as_completed(tasks), total=len(assets)):
            asset, ticker_data = await future
            if ticker_data is None:
                not_found.append(asset)
                continue
            
            close_data = prepare_data(ticker_data, asset).reset_index()
            
            if portfolio.empty:
                portfolio = close_data[['Date', asset]]
            elif asset not in portfolio.columns:
                portfolio = pd.merge(portfolio, close_data[['Date', asset]], on='Date', how='outer')
            
            # Random sleep to avoid rate limiting
            await asyncio.sleep(random.uniform(0.5, 1.0))

    portfolio = portfolio.set_index('Date')
    portfolio.index = pd.to_datetime(portfolio.index)
    portfolio = portfolio.round(3)
    if len(not_found):
        print(f"Not found assets: {len(set(not_found))}, {set(not_found)}")
    return portfolio

def prepare_data(data, ticker):
    data = data.drop(['open', 'high', 'low', 'volume'], axis=1)
    data = data.rename(columns={"close": ticker})
    return data.round(2)


def get_portfolio(assets, from_='3m', start_date=None, end_date=dt.now().date()):
    """
    Synchronous wrapper for the asynchronous get_portfolio_async function.
    """
    loop = asyncio.get_event_loop()
    portfolio = loop.run_until_complete(get_portfolio_async(assets, from_=from_, start_date=start_date, end_date=end_date))
    return portfolio



from earningspy.generators.finviz.main_func import (get_all_news, get_analyst_price_targets,
                              get_insider, get_news, get_stock)
from earningspy.generators.finviz.portfolio import Portfolio
from earningspy.generators.finviz.screener import Screener



import pytest
import numpy as np
from earningspy.generators.finviz.data import (
    _get_screener_data,
    get_by_earnings_date,
    get_by_tickers,
)

EXPECTED_COLUMNS = [
    'Ticker', 'Company', 'Sector', 'Industry', 'Country', 'Market Cap',
    'P/E', 'Fwd P/E', 'PEG', 'P/S', 'P/B', 'P/C', 'P/FCF', 'Dividend',
    'Payout Ratio', 'EPS', 'EPS next Q', 'EPS This Y', 'EPS Next Y',
    'EPS Past 5Y', 'EPS Next 5Y', 'Sales Past 5Y', 'Sales Q/Q',
    'Sales YoY TTM', 'EPS Q/Q', 'Sales', 'Income', 'EPS Surprise',
    'Revenue Surprise', 'Outstanding', 'Float', 'Float %', 'Insider Own',
    'Insider Trans', 'Inst Own', 'Inst Trans', 'Short Float', 'Short Ratio',
    'Short Interest', 'ROA', 'ROE', 'ROIC', 'Curr R', 'Quick R',
    'LTDebt/Eq', 'Debt/Eq', 'Gross M', 'Oper M', 'Profit M', 'Perf Week',
    'Perf Month', 'Perf Quart', 'Perf Half', 'Perf Year', 'Perf YTD',
    'Beta', 'ATR', 'Volatility W', 'Volatility M', 'SMA20', 'SMA50',
    'SMA200', 'RSI', 'Target Price', 'Book/sh', 'Cash/sh', 'Employees',
    'Optionable', 'Prev Close', 'Shortable', 'Recom', 'Avg Volume',
    'Rel Volume', 'Volume', 'Price', 'Change', 'Dividend TTM',
    'Dividend Ex Date', 'EPS YoY TTM', 'Enterprise Value', 'EV/EBITDA', 'EV/Sales',
    '52W_NORM', 'IS_S&P500', 'IS_RUSSELL', 'IS_NASDAQ', 'IS_DOW_JONES', 'IS_AMC',
    'IS_BMO', 'IS_USA', 'EARNINGS_DATE', 'FCF', 'EBITDA', 'EBIT', 'DATADATE'
]

MAX_ROW_NAN_RATIO = 0.30  # 30% NaNs allowed per row


def _assert_row_nan_threshold(data, max_ratio=MAX_ROW_NAN_RATIO, top_k=10):
    # NaN ratio per row across ALL columns
    row_nan_ratio = data.isna().mean(axis=1)

    bad = row_nan_ratio[row_nan_ratio > max_ratio]
    if not bad.empty:
        worst = bad.sort_values(ascending=False).head(top_k)
        msg_lines = [
            f"{idx}: {ratio:.2%} NaNs | Ticker={data.loc[idx, 'Ticker'] if 'Ticker' in data.columns else 'N/A'}"
            for idx, ratio in worst.items()
        ]
        raise AssertionError(
            f"Found {len(bad)} rows with NaN ratio > {max_ratio:.0%}. "
            f"Showing worst {min(top_k, len(bad))} rows:\n" + "\n".join(msg_lines)
        )


@pytest.mark.parametrize("filters", [
    'ind_solar',
])
def test_get_screener_data(filters):
    print(f"Testing filter: {filters}")
    data = _get_screener_data(filters=filters, order='marketcap')

    assert data.columns.tolist() == EXPECTED_COLUMNS, "DataFrame columns do not match expected columns"
    assert not data.empty, "DataFrame should not be empty"

    _assert_row_nan_threshold(data)


@pytest.mark.parametrize("scopes", [
    'next_week',
])
def test_get_by_earnings_date(scopes):
    print(f"Testing scope: {scopes}")
    data = get_by_earnings_date(scopes)

    assert data.columns.tolist() == EXPECTED_COLUMNS, "DataFrame columns do not match expected columns"
    assert not data.empty, "DataFrame should not be empty"

    _assert_row_nan_threshold(data)


def test_get_by_tickers():
    tickers = ['AAPL', 'MSFT', 'GOOGL']
    data = get_by_tickers(tickers)

    assert data.columns.tolist() == EXPECTED_COLUMNS, "DataFrame columns do not match expected columns"
    assert not data.empty, "DataFrame should not be empty"

    _assert_row_nan_threshold(data)


def create_table_string(table_list):
    """ Used to create a readable representation of a table. """

    col_size = [max(map(len, col)) for col in zip(*table_list)]
    format_str = " | ".join([f"{{:<{i}}}" for i in col_size])
    table_list.insert(1, ["-" * i for i in col_size])

    table_string = ""
    for item in table_list:
        table_string += format_str.format(*item) + "\n"

    return table_string



import datetime
import os
import time

import requests
from lxml import etree, html


def get_table(page_html: requests.Response, headers, rows=None, **kwargs):
    """ Private function used to return table data inside a list of dictionaries. """
    if isinstance(page_html, str):
        page_parsed = html.fromstring(page_html)
    else:
        page_parsed = html.fromstring(page_html.text)
    # When we call this method from Portfolio we don't fill the rows argument.
    # Conversely, we always fill the rows argument when we call this method from Screener.
    # Also, in the portfolio page, we don't need the last row - it's redundant.
    if rows is None:
        rows = -2  # We'll increment it later (-1) and use it to cut the last row

    data_sets = []
    # Select the HTML of the rows and append each column text to a list
    all_rows = [
        column.xpath("td//text()")
        for column in page_parsed.cssselect('tr[valign="top"]')
    ]

    # If rows is different from -2, this function is called from Screener
    if rows != -2:
        for row_number, row_data in enumerate(all_rows, 1):
            data_sets.append(dict(zip(headers, row_data)))
            if row_number == rows:  # If we have reached the required end
                break
    else:
        # Zip each row values to the headers and append them to data_sets
        [data_sets.append(dict(zip(headers, row))) for row in all_rows]

    return data_sets


def get_total_rows(page_content):
    """ Returns the total number of rows(results). """

    options=[('class="count-text whitespace-nowrap">#1 / ',' Total</div>'),('class="count-text">#1 / ',' Total</td>')]
    page_text = str(html.tostring(page_content))
    for option_beg,option_end in options:
        if option_beg in page_text:
            total_number = page_text.split(option_beg)[1].split(option_end)[0]
            try:
                return int(total_number)
            except ValueError:
                return 0
    return 0


def get_page_urls(page_content, rows, url):
    """ Returns a list containing all of the page URL addresses. """

    total_pages = int(
        [i.text.split("/")[1] for i in page_content.cssselect('option[value="1"]')][0]
    )
    urls = []

    for page_number in range(1, total_pages + 1):
        sequence = 1 + (page_number - 1) * 20

        if sequence - 20 <= rows < sequence:
            break
        urls.append(url + f"&r={str(sequence)}")

    return urls


def download_chart_image(page_content: requests.Response, **kwargs):
    """ Downloads a .png image of a chart into the "charts" folder. """
    file_name = f"{kwargs['URL'].split('t=')[1]}_{int(time.time())}.png"

    if not os.path.exists("charts"):
        os.mkdir("charts")

    with open(os.path.join("charts", file_name), "wb") as handle:
        handle.write(page_content.content)


def get_analyst_price_targets_for_export(
    ticker=None, page_content=None, last_ratings=5
):
    analyst_price_targets = []

    try:
        table = page_content.cssselect('table[class="fullview-ratings-outer"]')[0]
        ratings_list = [row.xpath("td//text()") for row in table]
        ratings_list = [
            [val for val in row if val != "\n"] for row in ratings_list
        ]  # remove new line entries

        headers = [
            "ticker",
            "date",
            "category",
            "analyst",
            "rating",
            "price_from",
            "price_to",
        ]  # header names
        count = 0

        for row in ratings_list:
            if count == last_ratings:
                break

            price_from, price_to = (
                0,
                0,
            )  # default values for len(row) == 4 , that is there is NO price information
            if len(row) == 5:
                strings = row[4].split("→")
                if len(strings) == 1:
                    price_to = (
                        strings[0].strip(" ").strip("$")
                    )  # if only ONE price is available then it is 'price_to' value
                else:
                    price_from = (
                        strings[0].strip(" ").strip("$")
                    )  # both '_from' & '_to' prices available
                    price_to = strings[1].strip(" ").strip("$")

            elements = [
                ticker,
                datetime.datetime.strptime(row[0], "%b-%d-%y").strftime("%Y-%m-%d"),
            ]
            elements.extend(row[1:3])
            elements.append(row[3].replace("→", "->"))
            elements.append(price_from)
            elements.append(price_to)
            data = dict(zip(headers, elements))
            analyst_price_targets.append(data)
            count += 1
    except Exception:
        pass

    return analyst_price_targets


def download_ticker_details(page_content: requests.Response, **kwargs):
    data = {}
    ticker = kwargs["URL"].split("=")[1]
    page_parsed = html.fromstring(page_content.text)

    all_rows = [
        row.xpath("td//text()")
        for row in page_parsed.cssselect('tr[class="table-dark-row"]')
    ]

    for row in all_rows:
        for column in range(0, 11):
            if column % 2 == 0:
                data[row[column]] = row[column + 1]

    if len(data) == 0:
        raise Exception(f"No data found for ticker: {ticker}")

    return {ticker: [data, get_analyst_price_targets_for_export(ticker, page_parsed)]}



import datetime
from decimal import Decimal, ROUND_HALF_UP
import pandas as pd
import numpy as np
from earningspy.generators.finviz.constants import (
    PERCENTAJE_COLUMNS,
    MONEY_COLUMNS,
    NUMERIC_COLUMNS,
    FINVIZ_DROP_COLUMNS
)

FINVIZ_URL = "https://finviz.com/screener.ashx?v=152&f={}{}&o={}"


def round_half_up(value):
    if value is None or pd.isna(value):
        return np.nan
    return int(Decimal(str(value)).quantize(0, rounding=ROUND_HALF_UP))


def _process_money_value(value):
    if isinstance(value, np.float64):
        return value
    if type(value) == float:
        return value
    elif type(value) == int:
        return float(value)
    else:
        if value.endswith('B'):
            value = float(value.strip('B'))
            value = value * 1000000000
            return value
        elif value.endswith('M'):
            value = float(value.strip('M'))
            value = value * 1000000
            return value
        elif value.endswith('K'):
            value = float(value.strip('K'))
            value = value * 1000
            return value
        elif value == '-':
            return float(0.0)
        else:
            return value



def _format_percent(value):
    if value is None or (isinstance(value, float) and np.isnan(value)):
        return np.nan

    # Handle strings like "3%", " 3.5 % ", "-"
    if isinstance(value, str):
        value = value.strip()
        if value in {"", "-", "—"}:
            return np.nan
        value = value.replace("%", "")
        try:
            value = float(value)
        except ValueError:
            return np.nan

    # At this point value should be numeric
    try:
        value = float(value)
    except Exception:
        return np.nan

    # If value looks like a percentage (>1), convert to decimal
    if abs(value) > 1:
        value = value / 100

    return value


def _convert_percent_columns(data): 

    for col in PERCENTAJE_COLUMNS:
        try:
            data.loc[col] = data.loc[col].apply(_format_percent)
        except Exception as e:
            print('Unable to transform this column: {} - {} - {}'.format(col, e, e.__class__))
    return data


def _process_money_columns(df):

    for col in MONEY_COLUMNS:
        if col in df.index:
            df.loc[col] = df.loc[col].apply(_process_money_value)
        
    return df


def _process_52_high(value):

    if isinstance(value, float):
        return value
    elif isinstance(value, str):
        high_low = value.split(' - ')
        if high_low[1] == '-':
            return float(0.0)
        elif high_low[1] != '-':
            return float(high_low[1])
    return np.nan


def _process_52_low(value):
    if isinstance(value, float):
        return value
    high_low = value.split(' - ')
    if high_low[0] == '-':
        return float(0.0)
    elif high_low[0] != '-':
        return float(high_low[0])
    return np.nan


def _process_52_high_low(data, drop=False):
    
    col_name = '52W Range'
    low_col_name = '52W Low'
    high_col_name = '52W High'
    
    if not col_name in data.index:
        print('No 52W Range field')
        return data 
    
    data.loc[low_col_name] = data.loc[col_name].apply(_process_52_low)
    data.loc[high_col_name] = data.loc[col_name].apply(_process_52_high)
    if drop:
        data.drop(col_name, inplace=True)
        
    return data


def _calculate_normalized_52w(row):
    """Calculates the normalized indicator for price within a 52-week range."""

    if isinstance(row['52W Range'], float):
        return row['52W Range']

    range52 = row['52W Range'].split('-')
    if not len(range52):
        return np.nan
    try:
        low_52w, high_52w = range52
        low_52w, high_52w = float(low_52w.strip()), float(high_52w.strip())

        midpoint = (high_52w + low_52w) / 2
        range_width = high_52w - low_52w
        normalized_indicator = (row['Price'] - midpoint) / (range_width / 2)
    except Exception as e:
        print(f"Error processing 52W Range {range52}: {e}")
        return np.nan
    else:
        return np.round(normalized_indicator, 4)


def _process_index(row, index): 

    if isinstance(row['Index'], int):
        return row['Index']
    indexes = row['Index'].replace(' ', '').split(',')
    if index in indexes:
        return 1

    return 0


def _process_earnings_time(row, time):

    if isinstance(row['Earnings'], int):
        return row['Earnings']
    earnings_time = row['Earnings'].split('/')
    if earnings_time == ['-']:
        return np.nan
    try:
        earnings_time = earnings_time[1]
    except IndexError:
        return np.nan
    else:
        if earnings_time == time:
            return 1
        else:
            return 0


def _process_free_cash_flow(row):
    try:
        market_cap = row['Market Cap']
        pfcf = row['P/FCF']

        # Guard against invalid inputs
        if pd.isna(market_cap) or pd.isna(pfcf):
            return np.nan
        if pfcf == 0:
            return np.nan

        value = market_cap / pfcf

        # Catch inf / -inf / nan
        if not np.isfinite(value):
            return np.nan

        return round_half_up(value)

    except Exception:
        print("Error trying to compute FCF")
        return np.nan


def _process_ebitda(row):
    try:
        ev = row['Enterprise Value']
        ev_ebitda = row['EV/EBITDA']

        # Guard against invalid inputs
        if pd.isna(ev) or pd.isna(ev_ebitda):
            return np.nan
        if ev_ebitda == 0:
            return np.nan

        value = ev / ev_ebitda

        # Catch inf / -inf / nan
        if not np.isfinite(value):
            return np.nan

        return round_half_up(value)

    except Exception:
        print("Error trying to compute ebitda")
        return np.nan
    

def _process_ebit(row):
    try:
        oper_margin = row['Oper M']
        sales = row['Sales']
        if pd.isna(oper_margin) or pd.isna(sales) or sales == 0:
            return np.nan
        value = oper_margin * sales
        return round_half_up(value) if np.isfinite(value) else np.nan
    except KeyError:
        print("Error trying to compute ebitda")
        raise
    except Exception:
        print("Error trying to compute ebit")
        return np.nan
    

def _process_ex_dividend(row):

    value = row['Dividend Ex Date']
    if isinstance(value, datetime.date):
        return value
    if pd.isna(value):
        return np.nan
    if value.strip() == '-':
        return np.nan
    try:
        date = pd.to_datetime(value, format="%m/%d/%Y").date()
    except:
        return np.nan
    else:
        return date


def _process_yes_columns(row, col_name):

    if row[col_name] == 'Yes':
        return 1
    return 0


def _process_country(row):
    value = row['Country']
    try:
        if value.strip().lower() == 'usa':
            return 1
    except Exception:
        return np.nan
        
    return 0


def _process_volume(value):
    if isinstance(value, float):
        return value
    elif isinstance(value, str):
        if value == '-':
            return 0.0
        value = value.replace(',','')
        return float(value)
    else:
        return value


def _process_numeric_columns(data):
    for col in NUMERIC_COLUMNS:
        try:
            data.loc[col] = data.loc[col].apply(_process_volume)
        except Exception as e:
            print('Unable to transform this column: {} - {}'.format(col, e))
            data.loc[col] = np.nan
    return data


def _process_report_date(row): 
    value = row['Earnings']
    date_list = value.split('/')
    if len(date_list) == 1 and type(date_list[0]) == str and date_list[0] != '-':
        date = date_list[0] + f" {datetime.datetime.now().year}"
        date = pd.to_datetime(date, format='%b %d %Y')
    elif len(date_list) == 2:
        date = date_list[0].split(' ')
        date = date_list[0] + f" {datetime.datetime.now().year}"
        date = pd.to_datetime(date, format='%b %d %Y')
    else:
        date = np.nan
    return date


def _process_remaning_columns(data):

    data = data.T
    data.loc[:,'52W_NORM'] = data.apply(lambda row: _calculate_normalized_52w(row), axis=1)

    # Index data
    data.loc[:,'IS_S&P500'] = data.apply(lambda row: _process_index(row, index='S&P500'), axis=1)
    data.loc[:,'IS_RUSSELL'] = data.apply(lambda row: _process_index(row, index='RUT'), axis=1)
    data.loc[:,'IS_NASDAQ'] = data.apply(lambda row: _process_index(row, index='NDX'), axis=1)
    data.loc[:,'IS_DOW_JONES'] = data.apply(lambda row: _process_index(row, index='DJIA'), axis=1)

    # After Market close or Before Market Opens
    data.loc[:,'IS_AMC'] = data.apply(lambda row: _process_earnings_time(row, time='a'), axis=1)
    data.loc[:,'IS_BMO'] = data.apply(lambda row: _process_earnings_time(row, time='b'), axis=1)

    # Dividend data
    data.loc[:,'Dividend Ex Date'] = data.apply(lambda row: _process_ex_dividend(row), axis=1)

    # Options data
    data.loc[:,'Optionable'] = data.apply(lambda row: _process_yes_columns(row, col_name='Optionable'), axis=1)
    data.loc[:,'Shortable'] = data.apply(lambda row: _process_yes_columns(row, col_name='Shortable'), axis=1)

    # Is US company and earnings date
    data.loc[:,'IS_USA'] = data.apply(lambda row: _process_country(row), axis=1)
    data.loc[:,'EARNINGS_DATE'] = data.apply(lambda row: _process_report_date(row), axis=1)

    # Value investing data
    data.loc[:, 'FCF'] = data.apply(lambda row: _process_free_cash_flow(row), axis=1)
    data.loc[:, 'EBITDA'] = data.apply(lambda row: _process_ebitda(row), axis=1)
    data.loc[:, 'EBIT'] = data.apply(lambda row: _process_ebit(row), axis=1)

    # Data fetched on this date 
    data.loc[:, 'DATADATE'] = pd.to_datetime(datetime.datetime.now().date())

    return data.T


def finviz_data_preprocessor(df):
    
    df = _convert_percent_columns(df)
    df = _process_money_columns(df)
    df = _process_numeric_columns(df)
    df = _process_52_high_low(df)
    df = _process_remaning_columns(df)
    df = df.drop(FINVIZ_DROP_COLUMNS, axis=0, errors='ignore')

    df = df.infer_objects()

    return df.T



from datetime import datetime

from lxml import etree

from earningspy.generators.finviz.helper_functions.request_functions import http_request_get
from earningspy.generators.finviz.helper_functions.scraper_functions import get_table

STOCK_URL = "https://finviz.com/quote.ashx"
NEWS_URL = "https://finviz.com/news.ashx"
CRYPTO_URL = "https://finviz.com/crypto_performance.ashx"
STOCK_PAGE = {}


def get_page(ticker):
    global STOCK_PAGE

    if ticker not in STOCK_PAGE:
        STOCK_PAGE[ticker], _ = http_request_get(
            url=STOCK_URL, payload={"t": ticker}, parse=True
        )


def get_stock(ticker):
    """
    Returns a dictionary containing stock data.

    :param ticker: stock symbol
    :type ticker: str
    :return dict
    """

    get_page(ticker)
    page_parsed = STOCK_PAGE[ticker]

    title = page_parsed.cssselect('table[class="fullview-title"]')[0]
    keys = ["Company", "Sector", "Industry", "Country"]
    fields = [f.text_content() for f in title.cssselect('a[class="tab-link"]')]
    data = dict(zip(keys, fields))

    company_link = title.cssselect('a[class="tab-link"]')[0].attrib["href"]
    data["Website"] = company_link if company_link.startswith("http") else None

    all_rows = [
        row.xpath("td//text()")
        for row in page_parsed.cssselect('tr[class="table-dark-row"]')
    ]

    for row in all_rows:
        for column in range(0, 11, 2):
            if row[column] == "EPS next Y" and "EPS next Y" in data.keys():
                data["EPS growth next Y"] = row[column + 1]
                continue
            elif row[column] == "Volatility":
                vols = row[column + 1].split()
                data["Volatility (Week)"] = vols[0]
                data["Volatility (Month)"] = vols[1]
                continue

            data[row[column]] = row[column + 1]

    return data


def get_insider(ticker):
    """
    Returns a list of dictionaries containing all recent insider transactions.

    :param ticker: stock symbol
    :return: list
    """

    get_page(ticker)
    page_parsed = STOCK_PAGE[ticker]
    outer_table = page_parsed.cssselect('table[class="body-table insider-trading-table"]')

    if len(outer_table) == 0:
        return []

    table = outer_table[0]
    headers = table[0].xpath("td//text()")

    data = [dict(zip(
        headers,
        [etree.tostring(elem, method="text", encoding="unicode") for elem in row]
    )) for row in table[1:]]

    return data


def get_news(ticker):
    """
    Returns a list of sets containing news headline and url

    :param ticker: stock symbol
    :return: list
    """

    get_page(ticker)
    page_parsed = STOCK_PAGE[ticker]
    news_table = page_parsed.cssselect('table[id="news-table"]')

    if len(news_table) == 0:
        return []

    rows = news_table[0].xpath("./tr[not(@id)]")

    results = []
    date = None
    for row in rows:
        raw_timestamp = row.xpath("./td")[0].xpath("text()")[0][0:-2]

        if len(raw_timestamp) > 8:
            parsed_timestamp = datetime.strptime(raw_timestamp, "%b-%d-%y %I:%M%p")
            date = parsed_timestamp.date()
        else:
            parsed_timestamp = datetime.strptime(raw_timestamp, "%I:%M%p").replace(
                year=date.year, month=date.month, day=date.day)

        results.append((
            parsed_timestamp.strftime("%Y-%m-%d %H:%M"),
            row.xpath("./td")[1].cssselect('a[class="tab-link-news"]')[0].xpath("text()")[0],
            row.xpath("./td")[1].cssselect('a[class="tab-link-news"]')[0].get("href"),
            row.xpath("./td")[1].cssselect('div[class="news-link-right"] span')[0].xpath("text()")[0][1:]
        ))

    return results


def get_all_news():
    """
    Returns a list of sets containing time, headline and url
    :return: list
    """

    page_parsed, _ = http_request_get(url=NEWS_URL, parse=True)
    all_dates = [
        row.text_content() for row in page_parsed.cssselect('td[class="nn-date"]')
    ]
    all_headlines = [
        row.text_content() for row in page_parsed.cssselect('a[class="nn-tab-link"]')
    ]
    all_links = [
        row.get("href") for row in page_parsed.cssselect('a[class="nn-tab-link"]')
    ]

    return list(zip(all_dates, all_headlines, all_links))


def get_crypto(pair):
    """

    :param pair: crypto pair
    :return: dictionary
    """

    page_parsed, _ = http_request_get(url=CRYPTO_URL, parse=True)
    page_html, _ = http_request_get(url=CRYPTO_URL, parse=False)
    crypto_headers = page_parsed.cssselect('tr[valign="middle"]')[0].xpath("td//text()")
    crypto_table_data = get_table(page_html, crypto_headers)

    return crypto_table_data[pair]


def get_analyst_price_targets(ticker, last_ratings=5):
    """
    Returns a list of dictionaries containing all analyst ratings and Price targets
     - if any of 'price_from' or 'price_to' are not available in the DATA, then those values are set to default 0
    :param ticker: stock symbol
    :param last_ratings: most recent ratings to pull
    :return: list
    """

    analyst_price_targets = []

    try:
        get_page(ticker)
        page_parsed = STOCK_PAGE[ticker]
        table = page_parsed.cssselect(
            'table[class="js-table-ratings fullview-ratings-outer"]'
        )[0]

        for row in table:
            rating = row.xpath("td//text()")
            rating = [val.replace("→", "->").replace("$", "") for val in rating if val != "\n"]
            rating[0] = datetime.strptime(rating[0], "%b-%d-%y").strftime("%Y-%m-%d")

            data = {
                "date": rating[0],
                "category": rating[1],
                "analyst": rating[2],
                "rating": rating[3],
            }
            if len(rating) == 5:
                if "->" in rating[4]:
                    rating.extend(rating[4].replace(" ", "").split("->"))
                    del rating[4]
                    data["target_from"] = float(rating[4])
                    data["target_to"] = float(rating[5])
                else:
                    data["target"] = float(rating[4])

            analyst_price_targets.append(data)
    except Exception as e:
        pass

    return analyst_price_targets[:last_ratings]
